@@ -57,9 +57,8 @@ import Distribution.Verbosity
 import qualified Distribution.Compat.CharParsing as P
 
 import Control.Monad ( msum )
-import Data.List ( stripPrefix, groupBy, partition )
+import Data.List ( stripPrefix, groupBy )
 import qualified Data.List.NonEmpty as NE
-import Data.Either ( partitionEithers )
 import System.FilePath as FilePath
          ( dropExtension, normalise, splitDirectories, joinPath, splitPath
          , hasTrailingPathSeparator )
@@ -154,7 +153,7 @@ readBuildTargets verbosity pkg targetStrs = do
 
     return btargets
 
-checkTargetExistsAsFile :: UserBuildTarget -> NoCallStackIO (UserBuildTarget, Bool)
+checkTargetExistsAsFile :: UserBuildTarget -> IO (UserBuildTarget, Bool)
 checkTargetExistsAsFile t = do
     fexists <- existsAsFile (fileComponentOfTarget t)
     return (t, fexists)
@@ -554,7 +553,7 @@ ex_cs =
 --
 
 data ComponentKind = LibKind | FLibKind | ExeKind | TestKind | BenchKind
-  deriving (Eq, Ord, Show)
+  deriving (Eq, Ord, Show, Enum, Bounded)
 
 componentKind :: ComponentName -> ComponentKind
 componentKind (CLibName   _) = LibKind
