@@ -35,20 +35,8 @@ import Distribution.Utils.Generic(safeLast)
 
 import qualified Data.List.NonEmpty as NE
 import qualified Data.Map as Map
-import qualified Data.Set as S
 import Control.Exception as Exception
-         ( Exception(toException), bracket, catches
-         , Handler(Handler), handleJust, IOException, SomeException )
-#ifndef mingw32_HOST_OS
-import Control.Exception as Exception
-         ( Exception(fromException) )
-#endif
-import System.Exit
-         ( ExitCode(..) )
-import Distribution.Compat.Exception
-         ( catchIO, catchExit )
-import Control.Monad
-         ( forM_, mapM )
+         ( bracket, catches, Handler(Handler), handleJust )
 import System.Directory
          ( getTemporaryDirectory, doesDirectoryExist, doesFileExist,
            createDirectoryIfMissing, removeFile, renameDirectory,
@@ -84,22 +72,19 @@ import Distribution.Client.Setup
          , filterTestFlags )
 import Distribution.Client.Config
          ( getCabalDir, defaultUserInstall )
-import Distribution.Client.Sandbox.Timestamp
-         ( withUpdateTimestamps )
-import Distribution.Client.Sandbox.Types
-         ( SandboxPackageInfo(..), UseSandbox(..), isUseSandbox
-         , whenUsingSandbox )
 import Distribution.Client.Tar (extractTarGzFile)
 import Distribution.Client.Types as Source
 import Distribution.Client.BuildReports.Types
          ( ReportLevel(..) )
 import Distribution.Client.SetupWrapper
          ( setupWrapper, SetupScriptOptions(..), defaultSetupScriptOptions )
+import Distribution.Client.BuildReports.Anonymous (showBuildReport)
 import qualified Distribution.Client.BuildReports.Anonymous as BuildReports
 import qualified Distribution.Client.BuildReports.Storage as BuildReports
          ( storeAnonymous, storeLocal, fromInstallPlan, fromPlanningFailure )
 import qualified Distribution.Client.InstallSymlink as InstallSymlink
-         ( OverwritePolicy(..), symlinkBinaries )
+         ( symlinkBinaries )
+import Distribution.Client.Types.OverwritePolicy (OverwritePolicy (..))
 import qualified Distribution.Client.Win32SelfUpgrade as Win32SelfUpgrade
 import qualified Distribution.Client.World as World
 import qualified Distribution.InstalledPackageInfo as Installed
@@ -134,8 +119,7 @@ import qualified Distribution.Simple.Setup as Cabal
          , registerCommand, RegisterFlags(..), emptyRegisterFlags
          , testCommand, TestFlags(..) )
 import Distribution.Simple.Utils
-         ( createDirectoryIfMissingVerbose, comparing
-         , writeFileAtomic )
+         ( createDirectoryIfMissingVerbose, writeFileAtomic )
 import Distribution.Simple.InstallDirs as InstallDirs
          ( PathTemplate, fromPathTemplate, toPathTemplate, substPathTemplate
          , initialPathTemplateEnv, installDirsTemplateEnv )
@@ -146,18 +130,18 @@ import Distribution.Package
          , Package(..), HasMungedPackageId(..), HasUnitId(..)
          , UnitId )
 import Distribution.Types.Dependency
-         ( thisPackageVersion )
+         ( Dependency (..), mainLibSet )
 import Distribution.Types.GivenComponent
          ( GivenComponent(..) )
-import Distribution.Pretty ( prettyShow )
 import Distribution.Types.PackageVersionConstraint
-         ( PackageVersionConstraint(..) )
+         ( PackageVersionConstraint(..), thisPackageVersionConstraint )
 import Distribution.Types.MungedPackageId
 import qualified Distribution.PackageDescription as PackageDescription
 import Distribution.PackageDescription
-         ( PackageDescription, GenericPackageDescription(..), Flag(..)
-         , FlagAssignment, mkFlagAssignment, unFlagAssignment
-         , showFlagValue, diffFlagAssignment, nullFlagAssignment )
+         ( PackageDescription, GenericPackageDescription(..) )
+import Distribution.Types.Flag
+         ( PackageFlag(..), FlagAssignment, mkFlagAssignment
+         , showFlagAssignment, diffFlagAssignment, nullFlagAssignment )
 import Distribution.PackageDescription.Configuration
          ( finalizePD )
 import Distribution.Version
@@ -167,11 +151,11 @@ import Distribution.Simple.Utils as Utils
          , withTempDirectory )
 import Distribution.Client.Utils
          ( determineNumJobs, logDirChange, mergeBy, MergeResult(..)
-         , tryCanonicalizePath, ProgressPhase(..), progressMessage )
+         , ProgressPhase(..), progressMessage )
 import Distribution.System
          ( Platform, OS(Windows), buildOS, buildPlatform )
 import Distribution.Verbosity as Verbosity
-         ( Verbosity, modifyVerbosity, normal, verbose )
+         ( modifyVerbosity, normal, verbose )
 import Distribution.Simple.BuildPaths ( exeExtension )
 
 import qualified Data.ByteString as BS
@@ -201,8 +185,6 @@ install
   -> Compiler
   -> Platform
   -> ProgramDb
-  -> UseSandbox
-  -> Maybe SandboxPackageInfo
   -> GlobalFlags
   -> ConfigFlags
   -> ConfigExFlags
@@ -212,7 +194,7 @@ install
   -> BenchmarkFlags
   -> [UserTarget]
   -> IO ()
-install verbosity packageDBs repos comp platform progdb useSandbox mSandboxPkgInfo
+install verbosity packageDBs repos comp platform progdb
   globalFlags configFlags configExFlags installFlags
   haddockFlags testFlags benchmarkFlags userTargets0 = do
 
@@ -222,7 +204,6 @@ install verbosity packageDBs repos comp platform progdb useSandbox mSandboxPkgIn
         ++ " (if you didn't type --root-cmd, comment out root-cmd"
         ++ " in your ~/.cabal/config file)"
     let userOrSandbox = fromFlag (configUserInstall configFlags)
-                     || isUseSandbox useSandbox
     unless userOrSandbox $
         warn verbosity $ "the --global flag is deprecated -- "
         ++ "it is generally considered a bad idea to install packages "
@@ -240,18 +221,12 @@ install verbosity packageDBs repos comp platform progdb useSandbox mSandboxPkgIn
             processInstallPlan verbosity args installContext installPlan
   where
     args :: InstallArgs
-    args = (packageDBs, repos, comp, platform, progdb, useSandbox,
-            mSandboxPkgInfo, globalFlags, configFlags, configExFlags,
+    args = (packageDBs, repos, comp, platform, progdb,
+            globalFlags, configFlags, configExFlags,
             installFlags, haddockFlags, testFlags, benchmarkFlags)
 
-    die'' message = die' verbosity (message ++ if isUseSandbox useSandbox
-                                   then installFailedInSandbox else [])
-    -- TODO: use a better error message, remove duplication.
-    installFailedInSandbox =
-      "\nNote: when using a sandbox, all packages are required to have "
-      ++ "consistent dependencies. "
-      ++ "Try reinstalling/unregistering the offending packages or "
-      ++ "recreating the sandbox."
+    die'' = die' verbosity
+
     logMsg message rest = debugNoWrap verbosity message >> rest
 
 -- TODO: Make InstallContext a proper data type with documented fields.
@@ -269,8 +244,6 @@ type InstallArgs = ( PackageDBStack
                    , Compiler
                    , Platform
                    , ProgramDb
-                   , UseSandbox
-                   , Maybe SandboxPackageInfo
                    , GlobalFlags
                    , ConfigFlags
                    , ConfigExFlags
@@ -283,14 +256,14 @@ type InstallArgs = ( PackageDBStack
 makeInstallContext :: Verbosity -> InstallArgs -> Maybe [UserTarget]
                       -> IO InstallContext
 makeInstallContext verbosity
-  (packageDBs, repoCtxt, comp, _, progdb,_,_,
+  (packageDBs, repoCtxt, comp, _, progdb,
    globalFlags, _, configExFlags, installFlags, _, _, _) mUserTargets = do
 
     let idxState = flagToMaybe (installIndexState installFlags)
 
-    installedPkgIndex <- getInstalledPackages verbosity comp packageDBs progdb
-    sourcePkgDb       <- getSourcePackagesAtIndexState verbosity repoCtxt idxState
-    pkgConfigDb       <- readPkgConfigDb      verbosity progdb
+    installedPkgIndex   <- getInstalledPackages verbosity comp packageDBs progdb
+    (sourcePkgDb, _, _) <- getSourcePackagesAtIndexState verbosity repoCtxt idxState Nothing
+    pkgConfigDb         <- readPkgConfigDb      verbosity progdb
 
     checkConfigExFlags verbosity installedPkgIndex
                        (packageIndex sourcePkgDb) configExFlags
@@ -321,7 +294,7 @@ makeInstallContext verbosity
 makeInstallPlan :: Verbosity -> InstallArgs -> InstallContext
                 -> IO (Progress String String SolverInstallPlan)
 makeInstallPlan verbosity
-  (_, _, comp, platform, _, _, mSandboxPkgInfo,
+  (_, _, comp, platform,_,
    _, configFlags, configExFlags, installFlags,
    _, _, _)
   (installedPkgIndex, sourcePkgDb, pkgConfigDb,
@@ -330,7 +303,7 @@ makeInstallPlan verbosity
     solver <- chooseSolver verbosity (fromFlag (configSolver configExFlags))
               (compilerInfo comp)
     notice verbosity "Resolving dependencies..."
-    return $ planPackages verbosity comp platform mSandboxPkgInfo solver
+    return $ planPackages verbosity comp platform solver
           configFlags configExFlags installFlags
           installedPkgIndex sourcePkgDb pkgConfigDb pkgSpecifiers
 
@@ -339,7 +312,7 @@ processInstallPlan :: Verbosity -> InstallArgs -> InstallContext
                    -> SolverInstallPlan
                    -> IO ()
 processInstallPlan verbosity
-  args@(_,_, _, _, _, _, _, _, configFlags, _, installFlags, _, _, _)
+  args@(_,_, _, _, _, _, configFlags, _, installFlags, _, _, _)
   (installedPkgIndex, sourcePkgDb, _,
    userTargets, pkgSpecifiers, _) installPlan0 = do
 
@@ -362,7 +335,6 @@ processInstallPlan verbosity
 planPackages :: Verbosity
              -> Compiler
              -> Platform
-             -> Maybe SandboxPackageInfo
              -> Solver
              -> ConfigFlags
              -> ConfigExFlags
@@ -372,7 +344,7 @@ planPackages :: Verbosity
              -> PkgConfigDb
              -> [PackageSpecifier UnresolvedSourcePackage]
              -> Progress String String SolverInstallPlan
-planPackages verbosity comp platform mSandboxPkgInfo solver
+planPackages verbosity comp platform solver
              configFlags configExFlags installFlags
              installedPkgIndex sourcePkgDb pkgConfigDb pkgSpecifiers =
 
@@ -445,8 +417,6 @@ planPackages verbosity comp platform mSandboxPkgInfo solver
             in LabeledPackageConstraint pc ConstraintSourceConfigFlagOrTarget
           | pkgSpecifier <- pkgSpecifiers ]
 
-      . maybe id applySandboxInstallPolicy mSandboxPkgInfo
-
       . (if reinstall then reinstallTargets else id)
 
         -- Don't solve for executables, the legacy install codepath
@@ -602,8 +572,8 @@ checkPrintPlan verbosity installed installPlan sourcePkgDb
   when offline $ do
     let pkgs = [ confPkgSource cpkg
                | InstallPlan.Configured cpkg <- InstallPlan.toList installPlan ]
-    notFetched <- fmap (map packageInfoId)
-                  . filterM (fmap isNothing . checkFetched . packageSource)
+    notFetched <- fmap (map packageId)
+                  . filterM (fmap isNothing . checkFetched . srcpkgSource)
                   $ pkgs
     unless (null notFetched) $
       die' verbosity $ "Can't download packages in offline mode. "
@@ -685,24 +655,26 @@ printPlan dryRun verbosity plan sourcePkgDb = case plan of
     showPkg (pkg, _) = prettyShow (packageId pkg) ++
                        showLatest (pkg)
 
-    showPkgAndReason (ReadyPackage pkg', pr) = prettyShow (packageId pkg') ++
-          showLatest pkg' ++
-          showFlagAssignment (nonDefaultFlags pkg') ++
-          showStanzas (confPkgStanzas pkg') ++
-          showDep pkg' ++
-          case pr of
-            NewPackage     -> " (new package)"
-            NewVersion _   -> " (new version)"
-            Reinstall _ cs -> " (reinstall)" ++ case cs of
+    showPkgAndReason (ReadyPackage pkg', pr) = unwords
+        [ prettyShow (packageId pkg')
+        , showLatest pkg'
+        , showFlagAssignment (nonDefaultFlags pkg')
+        , showStanzas (confPkgStanzas pkg')
+        , showDep pkg'
+        , case pr of
+            NewPackage     -> "(new package)"
+            NewVersion _   -> "(new version)"
+            Reinstall _ cs -> "(reinstall)" ++ case cs of
                 []   -> ""
-                diff -> " (changes: "  ++ intercalate ", " (map change diff)
+                diff -> "(changes: "  ++ intercalate ", " (map change diff)
                         ++ ")"
+        ]
 
     showLatest :: Package srcpkg => srcpkg -> String
     showLatest pkg = case mLatestVersion of
         Just latestVersion ->
             if packageVersion pkg < latestVersion
-            then (" (latest: " ++ prettyShow latestVersion ++ ")")
+            then ("(latest: " ++ prettyShow latestVersion ++ ")")
             else ""
         Nothing -> ""
       where
@@ -713,22 +685,19 @@ printPlan dryRun verbosity plan sourcePkgDb = case plan of
                            (packageIndex sourcePkgDb)
                            (packageName pkg)
 
-    toFlagAssignment :: [Flag] -> FlagAssignment
+    toFlagAssignment :: [PackageFlag] -> FlagAssignment
     toFlagAssignment =  mkFlagAssignment . map (\ f -> (flagName f, flagDefault f))
 
     nonDefaultFlags :: ConfiguredPackage loc -> FlagAssignment
     nonDefaultFlags cpkg =
       let defaultAssignment =
             toFlagAssignment
-             (genPackageFlags (SourcePackage.packageDescription $
+             (genPackageFlags (SourcePackage.srcpkgDescription $
                                confPkgSource cpkg))
       in  confPkgFlags cpkg `diffFlagAssignment` defaultAssignment
 
     showStanzas :: [OptionalStanza] -> String
-    showStanzas = concatMap ((" *" ++) . showStanza)
-
-    showFlagAssignment :: FlagAssignment -> String
-    showFlagAssignment = concatMap ((' ' :) . showFlagValue) . unFlagAssignment
+    showStanzas = unwords . map (("*" ++) . showStanza)
 
     change (OnlyInLeft pkgid)        = prettyShow pkgid ++ " removed"
     change (InBoth     pkgid pkgid') = prettyShow pkgid ++ " -> "
@@ -762,7 +731,7 @@ printPlan dryRun verbosity plan sourcePkgDb = case plan of
 reportPlanningFailure :: Verbosity -> InstallArgs -> InstallContext -> String
                       -> IO ()
 reportPlanningFailure verbosity
-  (_, _, comp, platform, _, _, _
+  (_, _, comp, platform, _
   ,_, configFlags, _, installFlags, _, _, _)
   (_, sourcePkgDb, _, _, pkgSpecifiers, _)
   message = do
@@ -791,7 +760,7 @@ reportPlanningFailure verbosity
     -- Save solver log
     case logFile of
       Nothing -> return ()
-      Just template -> forM_ pkgids $ \pkgid ->
+      Just template -> for_ pkgids $ \pkgid ->
         let env = initialPathTemplateEnv pkgid dummyIpid
                     (compilerInfo comp) platform
             path = fromPathTemplate $ substPathTemplate env template
@@ -842,18 +811,15 @@ postInstallActions :: Verbosity
                    -> BuildOutcomes
                    -> IO ()
 postInstallActions verbosity
-  (packageDBs, _, comp, platform, progdb, useSandbox, mSandboxPkgInfo
+  (packageDBs, _, comp, platform, progdb
   ,globalFlags, configFlags, _, installFlags, _, _, _)
   targets installPlan buildOutcomes = do
 
-  updateSandboxTimestampsFile verbosity useSandbox mSandboxPkgInfo
-                              comp platform installPlan buildOutcomes
-
   unless oneShot $
     World.insert verbosity worldFile
       --FIXME: does not handle flags
-      [ World.WorldPkgInfo dep mempty
-      | UserTargetNamed dep <- targets ]
+      [ World.WorldPkgInfo (Dependency pn vr mainLibSet) mempty
+      | UserTargetNamed (PackageVersionConstraint pn vr) <- targets ]
 
   let buildReports = BuildReports.fromInstallPlan platform (compilerId comp)
                                                   installPlan buildOutcomes
@@ -866,7 +832,7 @@ postInstallActions verbosity
   when (reportingLevel == DetailedReports) $
     storeDetailedBuildReports verbosity logsDir buildReports
 
-  regenerateHaddockIndex verbosity packageDBs comp platform progdb useSandbox
+  regenerateHaddockIndex verbosity packageDBs comp platform progdb
                          configFlags installFlags buildOutcomes
 
   symlinkBinaries verbosity platform comp configFlags installFlags
@@ -886,13 +852,13 @@ storeDetailedBuildReports verbosity logsDir reports = sequence_
   [ do dotCabal <- getCabalDir
        let logFileName = prettyShow (BuildReports.package report) <.> "log"
            logFile     = logsDir </> logFileName
-           reportsDir  = dotCabal </> "reports" </> remoteRepoName remoteRepo
+           reportsDir  = dotCabal </> "reports" </> unRepoName (remoteRepoName remoteRepo)
            reportFile  = reportsDir </> logFileName
 
        handleMissingLogFile $ do
          buildLog <- readFile logFile
          createDirectoryIfMissing True reportsDir -- FIXME
-         writeFile reportFile (show (BuildReports.show report, buildLog))
+         writeFile reportFile (show (showBuildReport report, buildLog))
 
   | (report, Just repo) <- reports
   , Just remoteRepo <- [maybeRepoRemote repo]
@@ -919,12 +885,11 @@ regenerateHaddockIndex :: Verbosity
                        -> Compiler
                        -> Platform
                        -> ProgramDb
-                       -> UseSandbox
                        -> ConfigFlags
                        -> InstallFlags
                        -> BuildOutcomes
                        -> IO ()
-regenerateHaddockIndex verbosity packageDBs comp platform progdb useSandbox
+regenerateHaddockIndex verbosity packageDBs comp platform progdb
                        configFlags installFlags buildOutcomes
   | haddockIndexFileIsRequested && shouldRegenerateHaddockIndex = do
 
@@ -952,8 +917,7 @@ regenerateHaddockIndex verbosity packageDBs comp platform progdb useSandbox
     -- installed. Since the index can be only per-user or per-sandbox (see
     -- #1337), we don't do it for global installs or special cases where we're
     -- installing into a specific db.
-    shouldRegenerateHaddockIndex = (isUseSandbox useSandbox || normalUserInstall)
-                                && someDocsWereInstalled buildOutcomes
+    shouldRegenerateHaddockIndex = normalUserInstall && someDocsWereInstalled buildOutcomes
       where
         someDocsWereInstalled = any installedDocs . Map.elems
         installedDocs (Right (BuildResult DocsOk _ _)) = True
@@ -987,7 +951,7 @@ symlinkBinaries :: Verbosity
 symlinkBinaries verbosity platform comp configFlags installFlags
                 plan buildOutcomes = do
   failed <- InstallSymlink.symlinkBinaries platform comp
-                                           InstallSymlink.NeverOverwrite
+                                           NeverOverwrite
                                            configFlags installFlags
                                            plan buildOutcomes
   case failed of
@@ -1051,33 +1015,6 @@ printBuildFailures verbosity buildOutcomes =
     onExitFailure _               = ""
 #endif
 
-
--- | If we're working inside a sandbox and some add-source deps were installed,
--- update the timestamps of those deps.
-updateSandboxTimestampsFile :: Verbosity -> UseSandbox -> Maybe SandboxPackageInfo
-                            -> Compiler -> Platform
-                            -> InstallPlan
-                            -> BuildOutcomes
-                            -> IO ()
-updateSandboxTimestampsFile verbosity (UseSandbox sandboxDir)
-                            (Just (SandboxPackageInfo _ _ _ allAddSourceDeps))
-                            comp platform installPlan buildOutcomes =
-  withUpdateTimestamps verbosity sandboxDir (compilerId comp) platform $ \_ -> do
-    let allInstalled = [ pkg
-                       | InstallPlan.Configured pkg
-                            <- InstallPlan.toList installPlan
-                       , case InstallPlan.lookupBuildOutcome pkg buildOutcomes of
-                           Just (Right _success) -> True
-                           _                     -> False
-                       ]
-        allSrcPkgs   = [ confPkgSource cpkg | cpkg <- allInstalled ]
-        allPaths     = [ pth | LocalUnpackedPackage pth
-                            <- map packageSource allSrcPkgs]
-    allPathsCanonical <- mapM tryCanonicalizePath allPaths
-    return $! filter (`S.member` allAddSourceDeps) allPathsCanonical
-
-updateSandboxTimestampsFile _ _ _ _ _ _ _ = return ()
-
 -- ------------------------------------------------------------
 -- * Actually do the installations
 -- ------------------------------------------------------------
@@ -1096,16 +1033,11 @@ performInstallations :: Verbosity
                      -> InstallPlan
                      -> IO BuildOutcomes
 performInstallations verbosity
-  (packageDBs, repoCtxt, comp, platform, progdb, useSandbox, _,
+  (packageDBs, repoCtxt, comp, platform, progdb,
    globalFlags, configFlags, configExFlags, installFlags,
    haddockFlags, testFlags, _)
   installedPkgIndex installPlan = do
 
-  -- With 'install -j' it can be a bit hard to tell whether a sandbox is used.
-  whenUsingSandbox useSandbox $ \sandboxDir ->
-    when parallelInstall $
-      notice verbosity $ "Notice: installing into a sandbox located at "
-                         ++ sandboxDir
   info verbosity $ "Number of threads used: " ++ (show numJobs) ++ "."
 
   jobControl   <- if parallelInstall then newParallelJobControl numJobs
@@ -1266,7 +1198,7 @@ installReadyPackage platform cinfo configFlags
     -- We generate the legacy constraints as well as the new style precise deps.
     -- In the end only one set gets passed to Setup.hs configure, depending on
     -- the Cabal version we are talking to.
-    configConstraints  = [ thisPackageVersion srcid
+    configConstraints  = [ thisPackageVersionConstraint srcid
                          | ConfiguredId
                              srcid
                              (Just
@@ -1472,7 +1404,7 @@ installUnpackedPackage verbosity installLock numJobs
             let packageDBs = interpretPackageDbFlags
                                 (fromFlag (configUserInstall configFlags))
                                 (configPackageDBs configFlags)
-            forM_ ipkgs' $ \ipkg' ->
+            for_ ipkgs' $ \ipkg' ->
                 registerPackage verbosity comp progdb
                                 packageDBs ipkg'
                                 defaultRegisterOptions
@@ -1549,7 +1481,7 @@ installUnpackedPackage verbosity installLock numJobs
           if is_dir
             -- Sort so that each prefix of the package
             -- configurations is well formed
-            then mapM (readPkgConf pkgConfDest) . sort . filter notHidden
+            then traverse (readPkgConf pkgConfDest) . sort . filter notHidden
                     =<< getDirectoryContents pkgConfDest
             else fmap (:[]) $ readPkgConf "." pkgConfDest
       else return []
