@@ -6,9 +6,8 @@
            , NondecreasingIndentation
            , MagicHash
   #-}
-{-# OPTIONS_GHC -Wno-name-shadowing #-}
 {-# OPTIONS_GHC -Wno-unused-matches #-}
-{-# OPTIONS_HADDOCK hide #-}
+{-# OPTIONS_HADDOCK not-home #-}
 
 -----------------------------------------------------------------------------
 -- |
@@ -28,11 +27,10 @@ module GHC.IO.Handle.Text (
         hWaitForInput, hGetChar, hGetLine, hGetContents, hPutChar, hPutStr,
         commitBuffer',       -- hack, see below
         hGetBuf, hGetBufSome, hGetBufNonBlocking, hPutBuf, hPutBufNonBlocking,
-        memcpy, hPutStrLn,
+        memcpy, hPutStrLn, hGetContents',
     ) where
 
 import GHC.IO
-import GHC.IO.FD
 import GHC.IO.Buffer
 import qualified GHC.IO.BufferedIO as Buffered
 import GHC.IO.Exception
@@ -46,8 +44,8 @@ import Foreign
 import Foreign.C
 
 import qualified Control.Exception as Exception
-import Data.Typeable
 import System.IO.Error
+import Data.Either (Either(..))
 import Data.Maybe
 
 import GHC.IORef
@@ -271,7 +269,7 @@ unpack !buf !r !w acc0
               -- Here, we are rather careful to only put an *evaluated* character
               -- in the output string. Due to pointer tagging, this allows the consumer
               -- to avoid ping-ponging between the actual consumer code and the thunk code
-#ifdef CHARBUF_UTF16
+#if defined(CHARBUF_UTF16)
               -- reverse-order decoding of UTF-16
               c2 <- peekElemOff pbuf i
               if (c2 < 0xdc00 || c2 > 0xdffff)
@@ -353,10 +351,10 @@ unpack_nl !buf !r !w acc0
 -- list returned by 'hGetContents' @hdl@.
 --
 -- Any operation that fails because a handle is closed,
--- also fails if a handle is semi-closed.  The only exception is 'hClose'.
--- A semi-closed handle becomes closed:
+-- also fails if a handle is semi-closed.  The only exception is
+-- 'System.IO.hClose'.  A semi-closed handle becomes closed:
 --
---  * if 'hClose' is applied to it;
+--  * if 'System.IO.hClose' is applied to it;
 --
 --  * if an I\/O error occurs when reading an item from the handle;
 --
@@ -453,6 +451,90 @@ getSomeCharacters handle_@Handle__{..} buf@Buffer{..} =
     _otherwise ->
       return buf
 
+-- -----------------------------------------------------------------------------
+-- hGetContents'
+
+-- We read everything into a list of CharBuffer chunks, and convert it lazily
+-- to a string, which minimizes memory usage.
+-- In the worst case, space usage is at most that of the complete String,
+-- as the chunks can be garbage collected progressively.
+-- For streaming consumers, space usage is at most that of the list of chunks.
+
+-- | The 'hGetContents'' operation reads all input on the given handle
+-- before returning it as a 'String' and closing the handle.
+--
+-- @since 4.15.0.0
+
+hGetContents' :: Handle -> IO String
+hGetContents' handle = do
+    es <- wantReadableHandle "hGetContents'" handle (strictRead handle)
+    case es of
+      Right s -> return s
+      Left e ->
+          case fromException e of
+            Just ioe -> throwIO (augmentIOError ioe "hGetContents'" handle)
+            Nothing -> throwIO e
+
+strictRead :: Handle -> Handle__ -> IO (Handle__, Either SomeException String)
+strictRead h handle_@Handle__{..} = do
+    cbuf <- readIORef haCharBuffer
+    cbufs <- strictReadLoop' handle_ [] cbuf
+    (handle_', me) <- hClose_help handle_
+    case me of
+      Just e -> return (handle_', Left e)
+      Nothing -> do
+        s <- lazyBuffersToString haInputNL cbufs ""
+        return (handle_', Right s)
+
+strictReadLoop :: Handle__ -> [CharBuffer] -> CharBuffer -> IO [CharBuffer]
+strictReadLoop handle_ cbufs cbuf0 = do
+    mcbuf <- Exception.catch
+        (do r <- readTextDevice handle_ cbuf0
+            return (Just r))
+        (\e -> if isEOFError e
+                  then return Nothing
+                  else throw e)
+    case mcbuf of
+      Nothing -> return (cbuf0 : cbufs)
+      Just cbuf1 -> strictReadLoop' handle_ cbufs cbuf1
+
+-- If 'cbuf' is full, allocate a new buffer.
+strictReadLoop' :: Handle__ -> [CharBuffer] -> CharBuffer -> IO [CharBuffer]
+strictReadLoop' handle_ cbufs cbuf
+    | isFullCharBuffer cbuf = do
+        cbuf' <- newCharBuffer dEFAULT_CHAR_BUFFER_SIZE ReadBuffer
+        strictReadLoop handle_ (cbuf : cbufs) cbuf'
+    | otherwise = strictReadLoop handle_ cbufs cbuf
+
+-- Lazily convert a list of buffers to a String. The buffers are
+-- in reverse order: the first buffer is the end of the String.
+lazyBuffersToString :: Newline -> [CharBuffer] -> String -> IO String
+lazyBuffersToString LF = loop where
+    loop [] s = return s
+    loop (Buffer{..} : cbufs) s = do
+        s' <- unsafeInterleaveIO (unpack bufRaw bufL bufR s)
+        loop cbufs s'
+lazyBuffersToString CRLF = loop '\0' where
+    loop before [] s = return s
+    loop before (Buffer{..} : cbufs) s
+        | bufL == bufR = loop before cbufs s  -- skip empty buffers
+        | otherwise = do
+            -- When a CRLF is broken across two buffers, we already have a newline
+            -- from decoding the LF, so we ignore the CR in the current buffer.
+            s1 <- if before == '\n'
+                     then return s
+                     else do
+                       -- We restore trailing CR not followed by LF.
+                       c <- peekCharBuf bufRaw (bufR - 1)
+                       if c == '\r'
+                          then return ('\r' : s)
+                          else return s
+            s2 <- unsafeInterleaveIO (do
+                (s2, _) <- unpack_nl bufRaw bufL bufR s1
+                return s2)
+            c0 <- peekCharBuf bufRaw bufL
+            loop c0 cbufs s2
+
 -- ---------------------------------------------------------------------------
 -- hPutChar
 
@@ -493,9 +575,9 @@ hPutcBuffered handle_@Handle__{..} c = do
                 LineBuffering -> True
                 _             -> False
 
-    putc buf@Buffer{ bufRaw=raw, bufR=w } c = do
+    putc buf@Buffer{ bufRaw=raw, bufR=w } c' = do
        debugIO ("putc: " ++ summaryBuffer buf)
-       w'  <- writeCharBuf raw w c
+       w'  <- writeCharBuf raw w c'
        return buf{ bufR = w' }
 
 -- ---------------------------------------------------------------------------
@@ -537,6 +619,7 @@ hPutStrLn handle str = hPutStr' handle str True
   -- overhead of a single putChar '\n', which is quite high now that we
   -- have to encode eagerly.
 
+{-# NOINLINE hPutStr' #-}
 hPutStr' :: Handle -> String -> Bool -> IO ()
 hPutStr' handle str add_nl =
   do
@@ -558,6 +641,7 @@ hPutChars :: Handle -> [Char] -> IO ()
 hPutChars _      [] = return ()
 hPutChars handle (c:cs) = hPutChar handle c >> hPutChars handle cs
 
+-- Buffer offset is always zero.
 getSpareBuffer :: Handle__ -> IO (BufferMode, CharBuffer)
 getSpareBuffer Handle__{haCharBuffer=ref,
                         haBuffers=spare_ref,
@@ -617,7 +701,6 @@ writeBlocks hdl line_buffered add_nl nl
 --
 -- Write the contents of the buffer 'buf' ('sz' bytes long, containing
 -- 'count' bytes of data) to handle (handle must be block or line buffered).
-
 commitBuffer
         :: Handle                       -- handle to commit to
         -> RawCharBuffer -> Int         -- address and size (in bytes) of buffer
@@ -629,9 +712,10 @@ commitBuffer
 commitBuffer hdl !raw !sz !count flush release =
   wantWritableHandle "commitBuffer" hdl $ \h_@Handle__{..} -> do
       debugIO ("commitBuffer: sz=" ++ show sz ++ ", count=" ++ show count
-            ++ ", flush=" ++ show flush ++ ", release=" ++ show release)
+            ++ ", flush=" ++ show flush ++ ", release=" ++ show release ++ ", handle=" ++ show hdl)
 
-      writeCharBuffer h_ Buffer{ bufRaw=raw, bufState=WriteBuffer,
+      -- Offset taken from handle
+      writeCharBuffer h_ Buffer{ bufRaw=raw, bufState=WriteBuffer, bufOffset=0,
                                  bufL=0, bufR=count, bufSize=sz }
 
       when flush $ flushByteWriteBuffer h_
@@ -644,6 +728,8 @@ commitBuffer hdl !raw !sz !count flush release =
                spare_bufs <- readIORef haBuffers
                writeIORef haBuffers (BufferListCons raw spare_bufs)
 
+      -- bb <- readIORef haByteBuffer
+      -- debugIO ("commitBuffer: buffer=" ++ summaryBuffer bb ++ ", handle=" ++ show hdl)
       return ()
 
 -- backwards compatibility; the text package uses this
@@ -655,7 +741,7 @@ commitBuffer' raw sz@(I# _) count@(I# _) flush release h_@Handle__{..}
             ++ ", flush=" ++ show flush ++ ", release=" ++ show release)
 
       let this_buf = Buffer{ bufRaw=raw, bufState=WriteBuffer,
-                             bufL=0, bufR=count, bufSize=sz }
+                             bufL=0, bufR=count, bufSize=sz, bufOffset=0 }
 
       writeCharBuffer h_ this_buf
 
@@ -683,7 +769,7 @@ commitBuffer' raw sz@(I# _) count@(I# _) flush release h_@Handle__{..}
 -- 'hPutBuf' ignores any text encoding that applies to the 'Handle',
 -- writing the bytes directly to the underlying file or device.
 --
--- 'hPutBuf' ignores the prevailing 'TextEncoding' and
+-- 'hPutBuf' ignores the prevailing 'System.IO.TextEncoding' and
 -- 'NewlineMode' on the 'Handle', and writes bytes directly.
 --
 -- This operation may fail with:
@@ -730,43 +816,80 @@ hPutBuf' handle ptr count can_block
              _line_or_no_buffering -> do flushWriteBuffer h_
           return r
 
+-- TODO: Possible optimisation:
+--       If we know that `w + count > size`, we should write both the
+--       handle buffer and the `ptr` in a single `writev()` syscall.
 bufWrite :: Handle__-> Ptr Word8 -> Int -> Bool -> IO Int
-bufWrite h_@Handle__{..} ptr count can_block =
-  seq count $ do  -- strictness hack
-  old_buf@Buffer{ bufRaw=old_raw, bufR=w, bufSize=size }
-     <- readIORef haByteBuffer
+bufWrite h_@Handle__{..} ptr !count can_block = do
+  -- Get buffer to determine size and free space in buffer
+  old_buf@Buffer{ bufR=w, bufSize=size }
+      <- readIORef haByteBuffer
 
-  -- enough room in handle buffer?
-  if (size - w > count)
-        -- There's enough room in the buffer:
-        -- just copy the data in and update bufR.
-        then do debugIO ("hPutBuf: copying to buffer, w=" ++ show w)
-                copyToRawBuffer old_raw w ptr count
-                writeIORef haByteBuffer old_buf{ bufR = w + count }
-                return count
+  -- There's no need to buffer if the incoming data is larger than
+  -- the handle buffer (`count >= size`).
+  -- Check if we can try to buffer the given chunk of data.
+  b <- if (count < size && count <= size - w)
+        then bufferChunk h_ old_buf ptr count
+        else do
+          -- The given data does not fit into the buffer.
+          -- Either because it's too large for the buffer
+          -- or the buffer is too full. Either way we need
+          -- to flush the buffered data first.
+          flushed_buf <- flushByteWriteBufferGiven h_ old_buf
+          if count < size
+              -- The data is small enough to be buffered.
+              then bufferChunk h_ flushed_buf ptr count
+              else do
+                let offset = bufOffset flushed_buf
+                !bytes <- if can_block
+                            then do writeChunk         h_ (castPtr ptr) offset count
+                            else writeChunkNonBlocking h_ (castPtr ptr) offset count
+                -- Update buffer with actual bytes written.
+                writeIORef haByteBuffer $! bufferAddOffset bytes flushed_buf
+                return bytes
+  debugIO "hPutBuf: done"
+  return b
 
-        -- else, we have to flush
-        else do debugIO "hPutBuf: flushing first"
-                old_buf' <- Buffered.flushWriteBuffer haDevice old_buf
-                        -- TODO: we should do a non-blocking flush here
-                writeIORef haByteBuffer old_buf'
-                -- if we can fit in the buffer, then just loop
-                if count < size
-                   then bufWrite h_ ptr count can_block
-                   else if can_block
-                           then do writeChunk h_ (castPtr ptr) count
-                                   return count
-                           else writeChunkNonBlocking h_ (castPtr ptr) count
+-- Flush the given buffer via the handle, return the flushed buffer
+flushByteWriteBufferGiven :: Handle__ -> Buffer Word8 -> IO (Buffer Word8)
+flushByteWriteBufferGiven h_@Handle__{..} bbuf = do
+  if (not (isEmptyBuffer bbuf))
+    then do
+      bbuf' <- Buffered.flushWriteBuffer haDevice bbuf
+      debugIO ("flushByteWriteBufferGiven: bbuf=" ++ summaryBuffer bbuf')
+      writeIORef haByteBuffer bbuf'
+      return bbuf'
+    else
+      return bbuf
 
-writeChunk :: Handle__ -> Ptr Word8 -> Int -> IO ()
-writeChunk h_@Handle__{..} ptr bytes
-  | Just fd <- cast haDevice  =  RawIO.write (fd::FD) ptr bytes
-  | otherwise = error "Todo: hPutBuf"
+-- Fill buffer and return bytes buffered/written.
+-- Flushes buffer if it's full after adding the data.
+bufferChunk :: Handle__ -> Buffer Word8 -> Ptr Word8 -> Int -> IO Int
+bufferChunk h_@Handle__{..} old_buf@Buffer{ bufRaw=raw, bufR=w, bufSize=size } ptr !count = do
+    debugIO ("hPutBuf: copying to buffer, w=" ++ show w)
+    copyToRawBuffer raw w ptr count
+    let copied_buf = old_buf{ bufR = w + count }
+    -- If the write filled the buffer completely, we need to flush,
+    -- to maintain the "INVARIANTS on Buffers" from
+    -- GHC.IO.Buffer.checkBuffer: "a write buffer is never full".
+    if isFullBuffer copied_buf
+      then do
+        -- TODO: we should do a non-blocking flush here
+        debugIO "hPutBuf: flushing full buffer after writing"
+        _ <- flushByteWriteBufferGiven h_ copied_buf
+        return ()
+      else do
+        writeIORef haByteBuffer copied_buf
+    return count
 
-writeChunkNonBlocking :: Handle__ -> Ptr Word8 -> Int -> IO Int
-writeChunkNonBlocking h_@Handle__{..} ptr bytes
-  | Just fd <- cast haDevice  =  RawIO.writeNonBlocking (fd::FD) ptr bytes
-  | otherwise = error "Todo: hPutBuf"
+writeChunk :: Handle__ -> Ptr Word8 -> Word64 -> Int -> IO Int
+writeChunk h_@Handle__{..} ptr offset bytes
+  = do RawIO.write haDevice ptr offset bytes
+       return bytes
+
+writeChunkNonBlocking :: Handle__ -> Ptr Word8 -> Word64 -> Int -> IO Int
+writeChunkNonBlocking h_@Handle__{..} ptr offset bytes
+  = RawIO.writeNonBlocking haDevice ptr offset bytes
 
 -- ---------------------------------------------------------------------------
 -- hGetBuf
@@ -783,21 +906,25 @@ writeChunkNonBlocking h_@Handle__{..} ptr bytes
 -- If the handle is a pipe or socket, and the writing end
 -- is closed, 'hGetBuf' will behave as if EOF was reached.
 --
--- 'hGetBuf' ignores the prevailing 'TextEncoding' and 'NewlineMode'
+-- 'hGetBuf' ignores the prevailing 'System.IO.TextEncoding' and 'NewlineMode'
 -- on the 'Handle', and reads bytes directly.
 
 hGetBuf :: Handle -> Ptr a -> Int -> IO Int
-hGetBuf h ptr count
+hGetBuf h !ptr count
   | count == 0 = return 0
   | count <  0 = illegalBufferSize h "hGetBuf" count
   | otherwise =
       wantReadableHandle_ "hGetBuf" h $ \ h_@Handle__{..} -> do
-         flushCharReadBuffer h_
-         buf@Buffer{ bufRaw=raw, bufR=w, bufL=r, bufSize=sz }
+          debugIO $ ":: hGetBuf - " ++ show h ++ " - " ++ show count
+          flushCharReadBuffer h_
+          buf@Buffer{ bufRaw=raw, bufR=w, bufL=r, bufSize=sz }
             <- readIORef haByteBuffer
-         if isEmptyBuffer buf
-            then bufReadEmpty    h_ buf (castPtr ptr) 0 count
-            else bufReadNonEmpty h_ buf (castPtr ptr) 0 count
+          debugIO ("hGetBuf: " ++ summaryBuffer buf)
+          res <- if isEmptyBuffer buf
+                    then bufReadEmpty    h_ buf (castPtr ptr) 0 count
+                    else bufReadNonEmpty h_ buf (castPtr ptr) 0 count
+          debugIO "** hGetBuf done."
+          return res
 
 -- small reads go through the buffer, large reads are satisfied by
 -- taking data first from the buffer and then direct from the file
@@ -805,9 +932,14 @@ hGetBuf h ptr count
 
 bufReadNonEmpty :: Handle__ -> Buffer Word8 -> Ptr Word8 -> Int -> Int -> IO Int
 bufReadNonEmpty h_@Handle__{..}
+                -- w for width, r for ... read ptr?
                 buf@Buffer{ bufRaw=raw, bufR=w, bufL=r, bufSize=sz }
                 ptr !so_far !count
  = do
+        debugIO ":: bufReadNonEmpty"
+        -- We use < instead of <= because for count == avail
+        -- we need to reset bufL and bufR to zero.
+        -- See also: INVARIANTS on Buffers
         let avail = w - r
         if (count < avail)
            then do
@@ -823,30 +955,47 @@ bufReadNonEmpty h_@Handle__{..}
             so_far' = so_far + avail
             ptr' = ptr `plusPtr` avail
 
-        if remaining == 0
+        debugIO ("bufReadNonEmpty: " ++ summaryBuffer buf' ++ " s:" ++ show so_far' ++ " r:" ++ show remaining)
+        b <- if remaining == 0
            then return so_far'
            else bufReadEmpty h_ buf' ptr' so_far' remaining
+        debugIO ":: bufReadNonEmpty - done"
+        return b
 
-
+-- We want to read more data, but the buffer is empty. (buffL == buffR == 0)
+-- See also Note [INVARIANTS on Buffers] in Buffer.hs
 bufReadEmpty :: Handle__ -> Buffer Word8 -> Ptr Word8 -> Int -> Int -> IO Int
 bufReadEmpty h_@Handle__{..}
-             buf@Buffer{ bufRaw=raw, bufR=w, bufL=r, bufSize=sz }
+             buf@Buffer{ bufRaw=raw, bufR=w, bufL=_r, bufSize=sz, bufOffset=bff }
              ptr so_far count
- | count > sz, Just fd <- cast haDevice = loop fd 0 count
+ | count > sz
+ = do
+        bytes_read <- loop haDevice 0 bff count
+        -- bytes_read includes so_far (content that was in the buffer)
+        -- but that is already accounted for in the old offset, so don't
+        -- count it twice.
+        let buf1 = bufferAddOffset (fromIntegral $ bytes_read - so_far) buf
+        writeIORef haByteBuffer buf1
+        debugIO ("bufReadEmpty1.1: " ++ summaryBuffer buf1 ++ " read:" ++ show bytes_read)
+        return bytes_read
  | otherwise = do
-     (r,buf') <- Buffered.fillReadBuffer haDevice buf
-     if r == 0
-        then return so_far
-        else do writeIORef haByteBuffer buf'
-                bufReadNonEmpty h_ buf' ptr so_far count
+        (r,buf') <- Buffered.fillReadBuffer haDevice buf
+        writeIORef haByteBuffer buf'
+        if r == 0 -- end of file reached
+            then return so_far
+            else bufReadNonEmpty h_ buf' ptr so_far count
  where
-  loop :: FD -> Int -> Int -> IO Int
-  loop fd off bytes | bytes <= 0 = return (so_far + off)
-  loop fd off bytes = do
-    r <- RawIO.read (fd::FD) (ptr `plusPtr` off) bytes
+  -- Read @bytes@ byte into ptr. Repeating the read until either zero
+  -- bytes where read, or we are done reading.
+  loop :: RawIO.RawIO dev => dev -> Int -> Word64 -> Int -> IO Int
+  loop dev delta off bytes | bytes <= 0 = return (so_far + delta)
+  loop dev delta off bytes = do
+    r <- RawIO.read dev (ptr `plusPtr` delta) off bytes
+    debugIO $ show ptr ++ " - loop read@" ++ show delta ++ ": " ++ show r
+    debugIO $ "next:" ++ show (delta + r) ++ " - left:" ++ show (bytes - r)
     if r == 0
-        then return (so_far + off)
-        else loop fd (off + r) (bytes - r)
+        then return (so_far + delta)
+        else loop dev (delta + r) (off + fromIntegral r) (bytes - r)
 
 -- ---------------------------------------------------------------------------
 -- hGetBufSome
@@ -865,20 +1014,23 @@ bufReadEmpty h_@Handle__{..}
 -- If the handle is a pipe or socket, and the writing end
 -- is closed, 'hGetBufSome' will behave as if EOF was reached.
 --
--- 'hGetBufSome' ignores the prevailing 'TextEncoding' and 'NewlineMode'
--- on the 'Handle', and reads bytes directly.
+-- 'hGetBufSome' ignores the prevailing 'System.IO.TextEncoding' and
+-- 'NewlineMode' on the 'Handle', and reads bytes directly.
 
 hGetBufSome :: Handle -> Ptr a -> Int -> IO Int
-hGetBufSome h ptr count
+hGetBufSome h !ptr count
   | count == 0 = return 0
   | count <  0 = illegalBufferSize h "hGetBufSome" count
   | otherwise =
       wantReadableHandle_ "hGetBufSome" h $ \ h_@Handle__{..} -> do
          flushCharReadBuffer h_
-         buf@Buffer{ bufSize=sz } <- readIORef haByteBuffer
+         buf@Buffer{ bufSize=sz, bufOffset=offset } <- readIORef haByteBuffer
          if isEmptyBuffer buf
             then case count > sz of  -- large read? optimize it with a little special case:
-                    True | Just fd <- haFD h_ -> do RawIO.read fd (castPtr ptr) count
+                    True -> do bytes <- RawIO.read haDevice (castPtr ptr) offset count
+                               -- Update buffer with actual bytes written.
+                               writeIORef haByteBuffer $! bufferAddOffset bytes buf
+                               return bytes
                     _ -> do (r,buf') <- Buffered.fillReadBuffer haDevice buf
                             if r == 0
                                then return 0
@@ -891,9 +1043,6 @@ hGetBufSome h ptr count
               let count' = min count (bufferElems buf)
               in bufReadNBNonEmpty h_ buf (castPtr ptr) 0 count'
 
-haFD :: Handle__ -> Maybe FD
-haFD h_@Handle__{..} = cast haDevice
-
 -- | 'hGetBufNonBlocking' @hdl buf count@ reads data from the handle @hdl@
 -- into the buffer @buf@ until either EOF is reached, or
 -- @count@ 8-bit bytes have been read, or there is no more data available
@@ -907,14 +1056,14 @@ haFD h_@Handle__{..} = cast haDevice
 -- If the handle is a pipe or socket, and the writing end
 -- is closed, 'hGetBufNonBlocking' will behave as if EOF was reached.
 --
--- 'hGetBufNonBlocking' ignores the prevailing 'TextEncoding' and
+-- 'hGetBufNonBlocking' ignores the prevailing 'System.IO.TextEncoding' and
 -- 'NewlineMode' on the 'Handle', and reads bytes directly.
 --
 -- NOTE: on Windows, this function does not work correctly; it
 -- behaves identically to 'hGetBuf'.
 
 hGetBufNonBlocking :: Handle -> Ptr a -> Int -> IO Int
-hGetBufNonBlocking h ptr count
+hGetBufNonBlocking h !ptr count
   | count == 0 = return 0
   | count <  0 = illegalBufferSize h "hGetBufNonBlocking" count
   | otherwise =
@@ -928,25 +1077,27 @@ hGetBufNonBlocking h ptr count
 
 bufReadNBEmpty :: Handle__ -> Buffer Word8 -> Ptr Word8 -> Int -> Int -> IO Int
 bufReadNBEmpty   h_@Handle__{..}
-                 buf@Buffer{ bufRaw=raw, bufR=w, bufL=r, bufSize=sz }
+                 buf@Buffer{ bufRaw=raw, bufR=w, bufL=_r, bufSize=sz
+                           , bufOffset=offset }
                  ptr so_far count
-  | count > sz,
-    Just fd <- cast haDevice = do
-       m <- RawIO.readNonBlocking (fd::FD) ptr count
+  | count > sz = do
+       m <- RawIO.readNonBlocking haDevice ptr offset count
        case m of
          Nothing -> return so_far
-         Just n  -> return (so_far + n)
+         Just n  -> do -- Update buffer with actual bytes written.
+                       writeIORef haByteBuffer $! bufferAddOffset n buf
+                       return (so_far + n)
 
  | otherwise = do
-     buf <- readIORef haByteBuffer
+    --  buf <- readIORef haByteBuffer
      (r,buf') <- Buffered.fillReadBuffer0 haDevice buf
      case r of
        Nothing -> return so_far
        Just 0  -> return so_far
-       Just r  -> do
+       Just r'  -> do
          writeIORef haByteBuffer buf'
-         bufReadNBNonEmpty h_ buf' ptr so_far (min count r)
-                          -- NOTE: new count is    min count r
+         bufReadNBNonEmpty h_ buf' ptr so_far (min count r')
+                          -- NOTE: new count is    min count r'
                           -- so we will just copy the contents of the
                           -- buffer in the recursive call, and not
                           -- loop again.
@@ -958,6 +1109,9 @@ bufReadNBNonEmpty h_@Handle__{..}
                   ptr so_far count
   = do
         let avail = w - r
+        -- We use < instead of <= because for count == avail
+        -- we need to reset bufL and bufR to zero.
+        -- See also [INVARIANTS on Buffers] in Buffer.hs
         if (count < avail)
            then do
                 copyFromRawBuffer ptr raw r count
