@@ -31,9 +31,7 @@ import {-# SOURCE #-} GHC.Exception( divZeroException, overflowException
                                    , underflowException
                                    , ratioZeroDenomException )
 
-#if defined(MIN_VERSION_integer_gmp)
-import GHC.Integer.GMP.Internals
-#endif
+import GHC.Num.BigNat (gcdInt,gcdWord)
 
 infixr 8  ^, ^^
 infixl 7  /, `quot`, `rem`, `div`, `mod`
@@ -326,7 +324,7 @@ instance  Real Int  where
 
 -- | @since 2.0.1
 instance  Integral Int  where
-    toInteger (I# i) = smallInteger i
+    toInteger (I# i) = IS i
 
     a `quot` b
      | b == 0                     = divZeroError
@@ -334,11 +332,8 @@ instance  Integral Int  where
                                                   -- in GHC.Int
      | otherwise                  =  a `quotInt` b
 
-    a `rem` b
+    !a `rem` b -- See Note [Special case of mod and rem is lazy]
      | b == 0                     = divZeroError
-       -- The quotRem CPU instruction fails for minBound `quotRem` -1,
-       -- but minBound `rem` -1 is well-defined (0). We therefore
-       -- special-case it.
      | b == (-1)                  = 0
      | otherwise                  =  a `remInt` b
 
@@ -348,11 +343,8 @@ instance  Integral Int  where
                                                   -- in GHC.Int
      | otherwise                  =  a `divInt` b
 
-    a `mod` b
+    !a `mod` b -- See Note [Special case of mod and rem is lazy]
      | b == 0                     = divZeroError
-       -- The divMod CPU instruction fails for minBound `divMod` -1,
-       -- but minBound `mod` -1 is well-defined (0). We therefore
-       -- special-case it.
      | b == (-1)                  = 0
      | otherwise                  =  a `modInt` b
 
@@ -368,6 +360,15 @@ instance  Integral Int  where
      | b == (-1) && a == minBound = (overflowError, 0)
      | otherwise                  =  a `divModInt` b
 
+{- Note [Special case of mod and rem is lazy]
+~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
+The `quotRem`/`divMod` CPU instruction fails for minBound `quotRem` -1, but
+minBound `rem` -1 is well-defined (0). We therefore special-case for `b == -1`,
+but not for `a == minBound` because of Note [Order of tests] in GHC.Int. But
+now we have to make sure the function stays strict in a, to guarantee unboxing.
+Hence the bang on a, see #18187.
+-}
+
 --------------------------------------------------------------
 -- Instances for @Word@
 --------------------------------------------------------------
@@ -398,7 +399,7 @@ instance Integral Word where
     divMod  (W# x#) y@(W# y#)
         | y /= 0                = (W# (x# `quotWord#` y#), W# (x# `remWord#` y#))
         | otherwise             = divZeroError
-    toInteger (W# x#)           = wordToInteger x#
+    toInteger (W# x#)           = integerFromWord# x#
 
 --------------------------------------------------------------
 -- Instances for Integer
@@ -410,19 +411,19 @@ instance  Real Integer  where
 
 -- | @since 4.8.0.0
 instance Real Natural where
-    toRational n = naturalToInteger n :% 1
+    toRational n = integerFromNatural n :% 1
 
 -- Note [Integer division constant folding]
 -- ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
 --
--- Constant folding of quot, rem, div, mod, divMod and quotRem for
--- Integer arguments depends crucially on inlining. Constant folding
--- rules defined in compiler/prelude/PrelRules.hs trigger for
--- quotInteger, remInteger and so on. So if calls to quot, rem and so on
--- were not inlined the rules would not fire. The rules would also not
--- fire if calls to quotInteger and so on were inlined, but this does not
--- happen because they are all marked with NOINLINE pragma - see documentation
--- of integer-gmp or integer-simple.
+-- Constant folding of quot, rem, div, mod, divMod and quotRem for Integer
+-- arguments depends crucially on inlining. Constant folding rules defined in
+-- GHC.Core.Opt.ConstantFold trigger for integerQuot, integerRem and so on.
+-- So if calls to quot, rem and so on were not inlined the rules would not fire.
+--
+-- The rules would also not fire if calls to integerQuot and so on were inlined,
+-- but this does not happen because they are all marked with NOINLINE pragma.
+
 
 -- | @since 2.0.1
 instance  Integral Integer where
@@ -430,41 +431,55 @@ instance  Integral Integer where
 
     {-# INLINE quot #-}
     _ `quot` 0 = divZeroError
-    n `quot` d = n `quotInteger` d
+    n `quot` d = n `integerQuot` d
 
     {-# INLINE rem #-}
     _ `rem` 0 = divZeroError
-    n `rem` d = n `remInteger` d
+    n `rem` d = n `integerRem` d
 
     {-# INLINE div #-}
     _ `div` 0 = divZeroError
-    n `div` d = n `divInteger` d
+    n `div` d = n `integerDiv` d
 
     {-# INLINE mod #-}
     _ `mod` 0 = divZeroError
-    n `mod` d = n `modInteger` d
+    n `mod` d = n `integerMod` d
 
     {-# INLINE divMod #-}
     _ `divMod` 0 = divZeroError
-    n `divMod` d = case n `divModInteger` d of
-                     (# x, y #) -> (x, y)
+    n `divMod` d = n `integerDivMod` d
 
     {-# INLINE quotRem #-}
     _ `quotRem` 0 = divZeroError
-    n `quotRem` d = case n `quotRemInteger` d of
-                      (# q, r #) -> (q, r)
+    n `quotRem` d = n `integerQuotRem` d
 
 -- | @since 4.8.0.0
 instance Integral Natural where
-    toInteger = naturalToInteger
+    toInteger = integerFromNatural
 
-    divMod = quotRemNatural
-    div    = quotNatural
-    mod    = remNatural
+    {-# INLINE quot #-}
+    _ `quot` 0 = divZeroError
+    n `quot` d = n `naturalQuot` d
 
-    quotRem = quotRemNatural
-    quot    = quotNatural
-    rem     = remNatural
+    {-# INLINE rem #-}
+    _ `rem` 0 = divZeroError
+    n `rem` d = n `naturalRem` d
+
+    {-# INLINE div #-}
+    _ `div` 0 = divZeroError
+    n `div` d = n `naturalQuot` d
+
+    {-# INLINE mod #-}
+    _ `mod` 0 = divZeroError
+    n `mod` d = n `naturalRem` d
+
+    {-# INLINE divMod #-}
+    _ `divMod` 0 = divZeroError
+    n `divMod` d = n `naturalQuotRem` d
+
+    {-# INLINE quotRem #-}
+    _ `quotRem` 0 = divZeroError
+    n `quotRem` d = n `naturalQuotRem` d
 
 --------------------------------------------------------------
 -- Instances for @Ratio@
@@ -565,16 +580,39 @@ fromIntegral = fromInteger . toInteger
     #-}
 
 {-# RULES
-"fromIntegral/Natural->Natural"  fromIntegral = id :: Natural -> Natural
-"fromIntegral/Natural->Integer"  fromIntegral = toInteger :: Natural->Integer
-"fromIntegral/Natural->Word"     fromIntegral = naturalToWord
+"fromIntegral/Natural->Natural"  fromIntegral = id            :: Natural -> Natural
+"fromIntegral/Natural->Integer"  fromIntegral = toInteger     :: Natural -> Integer
+"fromIntegral/Natural->Word"     fromIntegral = naturalToWord :: Natural -> Word
   #-}
 
+-- Don't forget the type signatures in the following rules! Without a type
+-- signature we ended up with the rule:
+--
+--  "fromIntegral/Int->Natural" forall a (d::Integral a).
+--        fromIntegral @a @Natural = naturalFromWord . fromIntegral @a d
+--
+-- but this rule is certainly not valid for every Integral type a!
+--
+-- This rule wraps any Integral input into Word's range. As a consequence,
+-- (2^64 :: Integer) was incorrectly wrapped to (0 :: Natural), see #19345.
+--
+-- A follow-up issue with this rule was that no underflow exception was raised
+-- for negative Int values (see #20066). We now use a naturalFromInt helper
+-- function to restore this behavior.
+
 {-# RULES
-"fromIntegral/Word->Natural"     fromIntegral = wordToNatural
-"fromIntegral/Int->Natural"     fromIntegral = intToNatural
+"fromIntegral/Word->Natural"    fromIntegral = naturalFromWord :: Word -> Natural
+"fromIntegral/Int->Natural"     fromIntegral = naturalFromInt  :: Int -> Natural
   #-}
 
+-- | Convert an Int into a Natural, throwing an underflow exception for negative
+-- values.
+naturalFromInt :: Int -> Natural
+{-# INLINE naturalFromInt #-}
+naturalFromInt x
+  | x < 0     = underflowError
+  | otherwise = naturalFromWord (fromIntegral x)
+
 -- | general coercion to fractional types
 realToFrac :: (Real a, Fractional b) => a -> b
 {-# NOINLINE [1] realToFrac #-}
@@ -763,31 +801,24 @@ lcm 0 _         =  0
 lcm x y         =  abs ((x `quot` (gcd x y)) * y)
 
 {-# RULES
-"gcd/Integer->Integer->Integer" gcd = gcdInteger
-"lcm/Integer->Integer->Integer" lcm = lcmInteger
-"gcd/Natural->Natural->Natural" gcd = gcdNatural
-"lcm/Natural->Natural->Natural" lcm = lcmNatural
+"gcd/Integer->Integer->Integer" gcd = integerGcd
+"lcm/Integer->Integer->Integer" lcm = integerLcm
+"gcd/Natural->Natural->Natural" gcd = naturalGcd
+"lcm/Natural->Natural->Natural" lcm = naturalLcm
  #-}
 
-#if defined(MIN_VERSION_integer_gmp)
--- GMP defines a more efficient Int# and Word# GCD
-
-gcdInt' :: Int -> Int -> Int
-gcdInt' (I# x) (I# y) = I# (gcdInt x y)
-
-gcdWord' :: Word -> Word -> Word
-gcdWord' (W# x) (W# y) = W# (gcdWord x y)
-
 {-# RULES
-"gcd/Int->Int->Int"             gcd = gcdInt'
-"gcd/Word->Word->Word"          gcd = gcdWord'
+"gcd/Int->Int->Int"             gcd = gcdInt
+"gcd/Word->Word->Word"          gcd = gcdWord
  #-}
 
-#endif
-
+-- See Note [Stable Unfolding for list producers] in GHC.Enum
+{-# INLINABLE integralEnumFrom #-}
 integralEnumFrom :: (Integral a, Bounded a) => a -> [a]
 integralEnumFrom n = map fromInteger [toInteger n .. toInteger (maxBound `asTypeOf` n)]
 
+-- See Note [Stable Unfolding for list producers] in GHC.Enum
+{-# INLINABLE integralEnumFromThen #-}
 integralEnumFromThen :: (Integral a, Bounded a) => a -> a -> [a]
 integralEnumFromThen n1 n2
   | i_n2 >= i_n1  = map fromInteger [i_n1, i_n2 .. toInteger (maxBound `asTypeOf` n1)]
@@ -796,9 +827,13 @@ integralEnumFromThen n1 n2
     i_n1 = toInteger n1
     i_n2 = toInteger n2
 
+-- See Note [Stable Unfolding for list producers] in GHC.Enum
+{-# INLINABLE integralEnumFromTo #-}
 integralEnumFromTo :: Integral a => a -> a -> [a]
 integralEnumFromTo n m = map fromInteger [toInteger n .. toInteger m]
 
+-- See Note [Stable Unfolding for list producers] in GHC.Enum
+{-# INLINABLE integralEnumFromThenTo #-}
 integralEnumFromThenTo :: Integral a => a -> a -> a -> [a]
 integralEnumFromThenTo n1 n2 m
   = map fromInteger [toInteger n1, toInteger n2 .. toInteger m]
