@@ -32,11 +32,13 @@
 #include "sm/NonMoving.h"
 #include "sm/NonMovingMark.h"
 #include "Profiling.h" // prof_arena
+#include "HeapAlloc.h"
 
 /* -----------------------------------------------------------------------------
    Forward decls.
    -------------------------------------------------------------------------- */
 
+int   isHeapAlloced       ( StgPtr p);
 static void  checkSmallBitmap    ( StgPtr payload, StgWord bitmap, uint32_t );
 static void  checkLargeBitmap    ( StgPtr payload, StgLargeBitmap*, uint32_t );
 static void  checkClosureShallow ( const StgClosure * );
@@ -45,6 +47,13 @@ static void  checkSTACK          (StgStack *stack);
 static W_    countNonMovingSegments ( struct NonmovingSegment *segs );
 static W_    countNonMovingHeap     ( struct NonmovingHeap *heap );
 
+/* -----------------------------------------------------------------------------
+   Debugging utility.
+   -------------------------------------------------------------------------- */
+
+// the HEAP_ALLOCED macro in function form. Useful for use in GDB or similar.
+int isHeapAlloced    ( StgPtr p) { return HEAP_ALLOCED(p); }
+
 /* -----------------------------------------------------------------------------
    Check stack sanity
    -------------------------------------------------------------------------- */
@@ -224,111 +233,6 @@ checkClosureProfSanity(const StgClosure *p)
 }
 #endif
 
-/* Note [Racing weak pointer evacuation]
- * ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
- * While debugging a GC crash (#18919) I noticed a spurious crash due to the
- * end-of-GC sanity check stumbling across a weak pointer with unevacuated key.
- * This can happen when two GC threads race to evacuate a weak pointer.
- * Specifically, we start out with a heap with a weak pointer reachable
- * from both a generation's weak pointer list and some other root-reachable
- * closure (e.g. a Just constructor):
- *
- *            O                      W
- *            ┌──────────┐           ┌──────────┐
- * Root ────→ │ Just     │     ╭───→ │ Weak#    │ ←─────── weak_ptr_list
- * Set        ├──────────┤     │     ├──────────┤
- *            │          │ ────╯     │ value    │ ─→ ...
- *            └──────────┘           │ key      │ ───╮    K
- *                                   │ ...      │    │    ┌──────────┐
- *                                   └──────────┘    ╰──→ │ ...      │
- *                                                        ├──────────┤
- *
- * The situation proceeds as follows:
- *
- * 1. Thread A initiates a GC, wakes up the GC worker threads, and starts
- *    evacuating roots.
- * 2. Thread A evacuates a weak pointer object O to location O'.
- * 3. Thread A fills the block where O' lives and pushes it to its
- *    work-stealing queue.
- * 4. Thread B steals the O' block and starts scavenging it.
- * 5. Thread A enters markWeakPtrList.
- * 6. Thread A starts evacuating W, resulting in Wb'.
- * 7. Thread B scavenges O', evacuating W', resulting in Wa'.
- * 8. Thread A and B are now racing to evacuate W. Only one will win the race
- *    (due to the CAS in copy_tag). Let the winning copy be called W'.
- * 9. W will be replaced by a forwarding pointer to the winning copy, W'.
- * 10. Whichever thread loses the race will retry evacuation, see
- *     that W has already been evacuated, and proceed as usual.
- * 10. W' will get added to weak_ptr_list by markWeakPtrList.
- * 11. Eventually W' will be scavenged.
- * 12. traverseWeakPtrList will see that W' has been scavenged and evacuate the
- *     its key.
- * 13. However, the copy that lost the race is not on `weak_ptr_list`
- *     and will therefore never get its `key` field scavenged (since
- *     `traverseWeakPtrList` will never see it).
- *
- * Now the heap looks like:
- *
- *            O'                     W (from-space)
- *            ┌──────────┐           ┌──────────┐
- * Root ────→ │ Just     │           │ Fwd-ptr  │ ───────────╮
- * Set        ├──────────┤           ├──────────┤            │
- *            │          │ ────╮     │ value    │ ─→ ...     │
- *            └──────────┘     │     │ key      │ ────────────────────────╮
- *                             │     │ ...      │            │            │
- *                             │     └──────────┘            │            │
- *                             │                             │            │
- *                             │     Wa'                     │            │
- *                             │     ┌──────────┐       ╭────╯            │
- *                             ╰───→ │ Weak#    │ ←─────┤                 │
- *                                   ├──────────┤       ╰─ weak_ptr_list  │
- *                                   │ value    │ ─→ ...                  │
- *                                   │ key      │ ───╮    K'              │
- *                                   │ ...      │    │    ┌──────────┐    │
- *                                   └──────────┘    ╰──→ │ ...      │    │
- *                                                        ├──────────┤    │
- *                                   Wb'                                  │
- *                                   ┌──────────┐                         │
- *                                   │ Weak#    │                         │
- *                                   ├──────────┤                         │
- *                                   │ value    │ ─→ ...                  │
- *                                   │ key      │ ───╮    K (from-space)  │
- *                                   │ ...      │    │    ┌──────────┐    │
- *                                   └──────────┘    ╰──→ │ 0xaaaaa  │ ←──╯
- *                                                        ├──────────┤
- *
- *
- * Without sanity checking this is fine; we have introduced a spurious copy of
- * W, Wb' into the heap but it is unreachable and therefore won't cause any
- * trouble. However, with sanity checking we may encounter this spurious copy
- * when walking the heap. Moreover, this copy was never added to weak_ptr_list,
- * meaning that its key field (along with the other fields mark as
- * non-pointers) will not get scavenged and will therefore point into
- * from-space.
- *
- * To avoid this checkClosure skips over the key field when it sees a weak
- * pointer. Note that all fields of Wb' *other* than the key field should be
- * valid, so we don't skip the closure entirely.
- *
- * We then do additional checking of all closures on the weak_ptr_lists, where
- * we *do* check `key`.
- */
-
-// Check validity of objects on weak_ptr_list.
-// See Note [Racing weak pointer evacuation].
-static void
-checkGenWeakPtrList( uint32_t g )
-{
-  for (StgWeak *w = generations[g].weak_ptr_list; w != NULL; w = w->link) {
-    ASSERT(LOOKS_LIKE_CLOSURE_PTR(w));
-    ASSERT(w->header.info == &stg_WEAK_info);
-    ASSERT(LOOKS_LIKE_CLOSURE_PTR(w->key));
-    ASSERT(LOOKS_LIKE_CLOSURE_PTR(w->value));
-    ASSERT(LOOKS_LIKE_CLOSURE_PTR(w->finalizer));
-    ASSERT(LOOKS_LIKE_CLOSURE_PTR(w->cfinalizers));
-  }
-}
-
 // Returns closure size in words
 StgOffset
 checkClosure( const StgClosure* p )
@@ -448,9 +352,12 @@ checkClosure( const StgClosure* p )
        * representative of the actual layout.
        */
       { StgWeak *w = (StgWeak *)p;
-        // N.B. Checking most of the fields here is not safe.
-        // See Note [Racing weak pointer evacuation] for why.
-        ASSERT(LOOKS_LIKE_CLOSURE_PTR(w->cfinalizers));
+        ASSERT(LOOKS_LIKE_CLOSURE_PTR(w->key));
+        ASSERT(LOOKS_LIKE_CLOSURE_PTR(w->value));
+        ASSERT(LOOKS_LIKE_CLOSURE_PTR(w->finalizer));
+        if (w->link) {
+          ASSERT(LOOKS_LIKE_CLOSURE_PTR(w->link));
+        }
         return sizeW_fromITBL(info);
       }
 
@@ -577,7 +484,7 @@ void checkHeapChain (bdescr *bd)
                 ASSERT( size >= MIN_PAYLOAD_SIZE + sizeofW(StgHeader) );
                 p += size;
 
-                /* skip over slop */
+                /* skip over slop, see Note [slop on the heap] */
                 while (p < bd->free &&
                        (*p < 0x1000 || !LOOKS_LIKE_INFO_PTR(*p))) { p++; }
             }
@@ -706,18 +613,8 @@ checkSTACK (StgStack *stack)
 void
 checkTSO(StgTSO *tso)
 {
-    StgTSO *next;
-    const StgInfoTable *info;
-
-    if (tso->what_next == ThreadKilled) {
-      /* The garbage collector doesn't bother following any pointers
-       * from dead threads, so don't check sanity here.
-       */
-      return;
-    }
-
-    next = tso->_link;
-    info = (const StgInfoTable*) tso->_link->header.info;
+    StgTSO *next = tso->_link;
+    const StgInfoTable *info = (const StgInfoTable*) tso->_link->header.info;
     load_load_barrier();
 
     ASSERT(next == END_TSO_QUEUE ||
@@ -730,6 +627,7 @@ checkTSO(StgTSO *tso)
         || tso->why_blocked == BlockedOnMVarRead
         || tso->why_blocked == BlockedOnBlackHole
         || tso->why_blocked == BlockedOnMsgThrowTo
+        || tso->why_blocked == BlockedOnIOCompletion
         || tso->why_blocked == NotBlocked
         ) {
         ASSERT(LOOKS_LIKE_CLOSURE_PTR(tso->block_info.closure));
@@ -738,9 +636,9 @@ checkTSO(StgTSO *tso)
     ASSERT(LOOKS_LIKE_CLOSURE_PTR(tso->bq));
     ASSERT(LOOKS_LIKE_CLOSURE_PTR(tso->blocked_exceptions));
     ASSERT(LOOKS_LIKE_CLOSURE_PTR(tso->stackobj));
-
-    // XXX are we checking the stack twice?
-    checkSTACK(tso->stackobj);
+    ASSERT(LOOKS_LIKE_CLOSURE_PTR(tso->global_link) &&
+            (tso->global_link == END_TSO_QUEUE ||
+             get_itbl((StgClosure*)tso->global_link)->type == TSO));
 }
 
 /*
@@ -750,16 +648,14 @@ checkTSO(StgTSO *tso)
 void
 checkGlobalTSOList (bool checkTSOs)
 {
-  StgTSO *tso;
-  uint32_t g;
-
-  for (g = 0; g < RtsFlags.GcFlags.generations; g++) {
-      for (tso=generations[g].threads; tso != END_TSO_QUEUE;
+  for (uint32_t g = 0; g < RtsFlags.GcFlags.generations; g++) {
+      for (StgTSO *tso = generations[g].threads; tso != END_TSO_QUEUE;
            tso = tso->global_link) {
           ASSERT(LOOKS_LIKE_CLOSURE_PTR(tso));
           ASSERT(get_itbl((StgClosure *)tso)->type == TSO);
-          if (checkTSOs)
+          if (checkTSOs) {
               checkTSO(tso);
+          }
 
           // If this TSO is dirty and in an old generation, it better
           // be on the mutable list.
@@ -768,22 +664,20 @@ checkGlobalTSOList (bool checkTSOs)
               tso->flags &= ~TSO_MARKED;
           }
 
-          {
-              StgStack *stack;
-              StgUnderflowFrame *frame;
-
-              stack = tso->stackobj;
-              while (1) {
-                  if (stack->dirty & STACK_DIRTY) {
-                      ASSERT(Bdescr((P_)stack)->gen_no == 0 || (stack->dirty & STACK_SANE));
-                      stack->dirty &= ~STACK_SANE;
-                  }
-                  frame = (StgUnderflowFrame*) (stack->stack + stack->stack_size
-                                                - sizeofW(StgUnderflowFrame));
-                  if (frame->info != &stg_stack_underflow_frame_info
-                      || frame->next_chunk == (StgStack*)END_TSO_QUEUE) break;
-                  stack = frame->next_chunk;
+          StgStack *stack = tso->stackobj;
+          while (1) {
+              if (stack->dirty & STACK_DIRTY) {
+                  ASSERT(Bdescr((P_)stack)->gen_no == 0 || (stack->dirty & STACK_SANE));
+                  stack->dirty &= ~STACK_SANE;
               }
+              StgUnderflowFrame *frame =
+                  (StgUnderflowFrame*) (stack->stack + stack->stack_size
+                          - sizeofW(StgUnderflowFrame));
+              if (frame->info != &stg_stack_underflow_frame_info
+                      || frame->next_chunk == (StgStack*)END_TSO_QUEUE) {
+                  break;
+              }
+              stack = frame->next_chunk;
           }
       }
   }
@@ -912,12 +806,17 @@ static void checkGeneration (generation *gen,
     ASSERT(countBlocks(gen->large_objects) == gen->n_large_blocks);
 
 #if defined(THREADED_RTS)
-    // heap sanity checking doesn't work with SMP for two reasons:
-    //   * we can't zero the slop (see Updates.h).  However, we can sanity-check
-    //     the heap after a major gc, because there is no slop.
+    // Note [heap sanity checking with SMP]
+    // ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
     //
-    //   * the nonmoving collector may be mutating its large object lists, unless we
-    //     were in fact called by the nonmoving collector.
+    // heap sanity checking doesn't work with SMP for two reasons:
+    //
+    //   * We can't zero the slop. However, we can sanity-check the heap after a
+    //     major gc, because there is no slop. See also Updates.h and Note
+    //     [zeroing slop when overwriting closures].
+    //
+    //   * The nonmoving collector may be mutating its large object lists,
+    //     unless we were in fact called by the nonmoving collector.
     if (!after_major_gc) return;
 #endif
 
@@ -953,12 +852,6 @@ static void checkGeneration (generation *gen,
         checkHeapChain(ws->scavd_list);
     }
 
-    // Check weak pointer lists
-    // See Note [Racing weak pointer evacuation].
-    for (uint32_t g = 0; g < RtsFlags.GcFlags.generations; g++) {
-      checkGenWeakPtrList(g);
-    }
-
     checkLargeObjects(gen->large_objects);
     checkCompactObjects(gen->compact_objects);
 }
